/*
 * piTclDlp.c --
 *
 *	Implementation of the high-level DLP functionality.
 *
 * Copyright (C) 1998 Andreas Kupries (a.kupries@westend.com)
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL I BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF THIS
 * SOFTWARE AND ITS DOCUMENTATION, EVEN IF I HAVE BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * I SPECIFICALLY DISCLAIM ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND
 * I HAVE NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 * ENHANCEMENTS, OR MODIFICATIONS.
 *
 * CVS: $Id: dlp.c,v 1.2 1999/03/27 15:51:49 aku Exp $
 */


#include "internal.h"


/* Macro to check wether the calling dlp object is connected to its
 * pilot. Tries to connect if not, see 'DlpActivate' for details.
 * Called by all subcommands before doing anything with the connection.
 */

#define CHK_ACTIVE(dlp,cmd,ip) \
if (DlpActivate (dlp,cmd,ip,1) != TCL_OK) { \
     return TCL_ERROR; \
}



/*
 * Forward declarations of internal procedures.
 */

/*
 * Procedures implementing the object commands.
 */

static void
PiTclDeleteObjectCmd _ANSI_ARGS_ ((ClientData cd));


static int
PiTclDlpObjectCmd _ANSI_ARGS_ ((ClientData     cd,
				Tcl_Interp*    interp,
				int            objc,
				Tcl_Obj* CONST objv []));


/* Procedures implementing the various subcommands
 */

static int
DlpActivate _ANSI_ARGS_ ((PiTclDlpInstance* dlp,
			  char*             cmd,
			  Tcl_Interp*       interp,
			  int               argc));

static int
DlpDelete _ANSI_ARGS_ ((PiTclDlpInstance* dlp,
			char*             cmd,
			Tcl_Interp*       interp,
			int               argc));

static int
DlpSession _ANSI_ARGS_ ((PiTclDlpInstance* dlp,
			 char*             cmd,
			 Tcl_Interp*       interp,
			 int               objc,
			 Tcl_Obj* CONST    objv []));

static int
DlpSystem _ANSI_ARGS_ ((PiTclDlpInstance* dlp,
			char*             cmd,
			Tcl_Interp*       interp,
			int               objc,
			Tcl_Obj* CONST    objv []));

static int
DlpTickle _ANSI_ARGS_ ((PiTclDlpInstance* dlp,
			char*             cmd,
			Tcl_Interp*       interp,
			int               objc,
			Tcl_Obj* CONST    objv []));

static int
DlpDb _ANSI_ARGS_ ((PiTclDlpInstance* dlp,
		    char*             cmd,
		    Tcl_Interp*       interp,
		    int               objc, 
		    Tcl_Obj* CONST    objv []));

/* secondary subcommands.
 */

static int
SystemCard _ANSI_ARGS_ ((PiTclDlpInstance* dlp,
			 char*             cmd,
			 Tcl_Interp*       interp,
			 int               objc,
			 Tcl_Obj* CONST    objv []));

static int
SystemSet _ANSI_ARGS_ ((PiTclDlpInstance* dlp,
			char*             cmd,
			Tcl_Interp*       interp,
			int               objc,
			Tcl_Obj* CONST    objv []));

static int
SystemGet _ANSI_ARGS_ ((PiTclDlpInstance*  dlp,
			char*              cmd,
			Tcl_Interp*        interp,
			int                objc,
			Tcl_Obj* CONST     objv []));

static int
DatabaseCreate _ANSI_ARGS_ ((PiTclDlpInstance*   dlp,
			     char*               cmd,
			     Tcl_Interp*         interp,
			     int                 objc, 
			     Tcl_Obj* CONST      objv []));
static int
DatabaseDelete _ANSI_ARGS_ ((PiTclDlpInstance*   dlp,
			     char*               cmd,
			     Tcl_Interp*         interp,
			     int                 objc, 
			     Tcl_Obj* CONST      objv []));
static int
DatabaseOpen _ANSI_ARGS_ ((PiTclDlpInstance*   dlp,
			   char*               cmd,
			   Tcl_Interp*         interp,
			   int                 objc, 
			   Tcl_Obj* CONST      objv []));

static int
DatabaseList _ANSI_ARGS_ ((PiTclDlpInstance*   dlp,
			   char*               cmd,
			   Tcl_Interp*         interp));

static int
DatabaseInfo _ANSI_ARGS_ ((PiTclDlpInstance*   dlp,
			   char*               cmd,
			   Tcl_Interp*         interp,
			   char*               dbName));

static int
ProcessCreateOptions _ANSI_ARGS_ ((PiTclDlpInstance*      dlp,
				   char*                  cmd,
				   Tcl_Interp*            interp,
				   int                    objc,
				   Tcl_Obj* CONST         objv [],
				   int*                   card,
				   int*                   flags,
				   int*                   version,
				   PiTclPackInfo** pack));

static int
ProcessOpenOptions _ANSI_ARGS_ ((PiTclDlpInstance*      dlp,
				 char*                  cmd,
				 Tcl_Interp*            interp,
				 int                    objc,
				 Tcl_Obj* CONST         objv [],
				 int*                   mode,
				 PiTclPackInfo** pack));

static int
PiTclDlpCreate _ANSI_ARGS_ ((Tcl_Interp*    interp,
			     PiTclInfo*     info,
			     char*          cmd,
			     int            objc,
			     Tcl_Obj* CONST objv []));

static int
PiTclDlpConvert _ANSI_ARGS_ ((Tcl_Interp*    interp,
			      PiTclInfo*     info,
			      char*          cmd,
			      int            objc,
			      Tcl_Obj* CONST objv []));

static int
PiTclDlpMap _ANSI_ARGS_ ((Tcl_Interp*    interp,
			  PiTclInfo*     info,
			  char*          cmd,
			  int            objc,
			  Tcl_Obj* CONST objv []));


/*
 *------------------------------------------------------*
 *
 *	PiTclDeleteObjectCmd --
 *
 *	------------------------------------------------*
 *	This procedure cleans up any state left over from
 *	a deleted DLP object.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Releases memory acquired in 'CreateDlpObject'.
 *
 *	Result:
 *		Nothing.
 *
 *------------------------------------------------------*
 */

static void
PiTclDeleteObjectCmd (cd)
ClientData cd;
{
  PiTclDlpInstance* dlp = (PiTclDlpInstance*) cd;

  /* Close the connection
   */

  if ((dlp->acceptDone) && (dlp->inSession)) {
    /* Close a running session, destroy the local cache
     * of database objects.
     */

    PiTclDlpDbCloseAll (dlp, NULL, dlp->interp, PITCL_DB_DESTROY);
    dlp_EndOfSync (dlp->sd, dlpEndCodeNormal);
  }

  if (dlp->sd >= 0) {
    pi_close (dlp->sd);
  }

  PiTclReleaseDlpObject (dlp);
}

/*
 *------------------------------------------------------*
 *
 *	PiTclDlpObjectCmd --
 *
 *	------------------------------------------------*
 *	This procedure realizes the command of DLP objects.
 *	See the manpages for details on what it does.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
PiTclDlpObjectCmd (cd, interp, objc, objv)
ClientData     cd;
Tcl_Interp*    interp;
int            objc;
Tcl_Obj* CONST objv [];
{
  /* Convert subcommand to easy-to-use index. This converts the
   * given Tcl_Obj too, caching the conversion result. Later
   * accesses are done in constant time :-)
   */

  int               pindex;
  PiTclDlpInstance* dlp = (PiTclDlpInstance*) cd;
  char*             cmd = ARG (0);

  static char *subcmd[] = {
    "activate", "db", "delete", "isactive", "session", "system", "tickle", NULL
  };
  enum subcmd {
    DLP_ACTIVATE, DLP_DB, DLP_DELETE, DLP_ISACT, DLP_SESSION, DLP_SYSTEM, DLP_TICKLE
  };	  

  DLP_CHECKARGS (objc < 2, interp, cmd, " <subcommand>...");

  if (Tcl_GetIndexFromObj(interp, objv [1], subcmd, "subcommand", 0,
			  &pindex) != TCL_OK) {
    return TCL_ERROR;
  }

  /* Now dispatch to the appropriate subcommands
   */

  switch (pindex) {
  case DLP_ISACT:
    DLP_CHECKARGS (objc != 2, interp, cmd, " isactive");
    Tcl_SetObjResult (interp, Tcl_NewIntObj (dlp->acceptDone));
    return TCL_OK;

  case DLP_ACTIVATE:
    return DlpActivate (dlp, cmd, interp, objc-1);

  case DLP_DB:
    return DlpDb (dlp, cmd, interp, objc-1, objv+1);

  case DLP_DELETE:
    return DlpDelete (dlp, cmd, interp, objc-1);

  case DLP_SESSION:
    return DlpSession (dlp, cmd, interp, objc-1, objv+1);

  case DLP_SYSTEM:
    return DlpSystem (dlp, cmd, interp, objc-1, objv+1);

  case DLP_TICKLE:
    return DlpTickle (dlp, cmd, interp, objc-1, objv+1);

  default:
    /* impossible */
    return TCL_ERROR;
  }

  /* We should not come to this place */

  return TCL_ERROR;
}

/*
 *------------------------------------------------------*
 *
 *	PiTclDlpCmd --
 *
 *	------------------------------------------------*
 *	This procedure realizes the 'piDlp' command.
 *	See the manpages for details on what it does.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

EXTERN int
PiTclDlpCmd (cd, interp, objc, objv)
ClientData     cd;		/* Reference to global data structure. */
Tcl_Interp*    interp;		/* Current interpreter. */
int            objc;		/* Number of arguments. */
Tcl_Obj* CONST objv [];		/* Argument objects. */
{
  PiTclInfo*   info = (PiTclInfo*) cd;

  /* Accepted syntax:
   *
   * piDlp create <objectName> ?<device>?
   * piDlp converter ...
   * piDlp map ...
   *
   * A missing device specification will cause the usage of environment
   * variable PILOTPORT.
   */

  static char *subcmd[] = {
    "converter", "create", "map", NULL
  };
  enum subcmd {
    PIDLP_CVT, PIDLP_CRT, PIDLP_MAP
  };

  int pindex;

  DLP_CHECKARGS (objc < 3, interp, "piDlp <subcommand>...", "");

  if (Tcl_GetIndexFromObj(interp, objv [1], subcmd, "subcommand", 0,
			  &pindex) != TCL_OK) {
    return TCL_ERROR;
  }

  switch (pindex) {

  case PIDLP_CVT:
    return PiTclDlpConvert (interp, info, ARG (0), objc-1, objv+1);

  case PIDLP_CRT:
    return PiTclDlpCreate  (interp, info, ARG (0), objc-1, objv+1);

  case PIDLP_MAP:
    return PiTclDlpMap     (interp, info, ARG (0), objc-1, objv+1);

  default:
    return TCL_ERROR;
  }

  return TCL_ERROR;
}

/*
 *------------------------------------------------------*
 *
 *	PiTclDlpCreate --
 *
 *	------------------------------------------------*
 *	This procedure realizes the 'create' subcommand
 *	of the dlp generator. See the manpages for
 *	details on what it does.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
PiTclDlpCreate (interp, info, cmd, objc, objv)
Tcl_Interp* interp;
PiTclInfo*  info;
char*       cmd;
int         objc;
Tcl_Obj* CONST objv [];
{
  char*        dev;
  PiTclDlpInstance* obj;

  /* (piDlp) create object ?device? */

  DLP_CHECKARGS ((objc < 2) || (objc > 3),
		 interp, cmd, "create object ?device?");

  if (objc == 3) {
    dev = ARG (2);
  } else {
    dev = getenv("PILOTPORT");

    if (dev == (char*) NULL) {
      /* Device neither given as argument nor by environment.
       * Use hardcoded fallback value (same as in pilot-link)
       */

      dev = "/dev/pilot";
    }
  }

  if (PiTclDeviceIsInUse (info, dev)) {
    Tcl_AppendResult (interp,
		      "piDlp: device already used",
		      (char*) NULL);
    return TCL_ERROR;
  }

  obj = PiTclCreateDlpObject (info, dev);

  if (obj == (PiTclDlpInstance*) NULL) {
        Tcl_AppendResult (interp,
		      "piDlp: could not create object",
		      (char*) NULL);
    return TCL_ERROR;
  }

  obj->interp = interp;
  obj->cmd    = Tcl_CreateObjCommand (interp, ARG (1),
				      &PiTclDlpObjectCmd,
				      (ClientData) obj,
				      &PiTclDeleteObjectCmd);

  /* One more reference to the global datastructures.
   */

  info->refCount ++;

  Tcl_SetObjResult (interp, objv [1]);
  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	PiTclDlpConvert --
 *
 *	------------------------------------------------*
 *	This procedure realizes the 'converter' subcommand
 *	of the dlp generator. See the manpages for
 *	details on what it does.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
PiTclDlpConvert (interp, info, cmd, objc, objv)
Tcl_Interp* interp;
PiTclInfo*  info;
char*       cmd;
int         objc;
Tcl_Obj* CONST objv [];
{
  /* (piDlp) converter create    <name> ?-option value...?
   * (piDlp) converter configure <name> ?-option value...?
   * (piDlp) converter delete    <name>
   *
   * (piDlp) converter configure <name> -option
   * (piDlp) converter configure <name>
   * (piDlp) converter cget      <name> -option
   *
   * (piDlp) converter list
   */

  static char *subcmd[] = {
    "cget", "configure", "create", "delete", "list", NULL
  };
  enum subcmd {
    CVT_CGET, CVT_CONF, CVT_CREATE, CVT_DELETE, CVT_LIST
  };

  PiTclPackInfo* p;
  int                   pindex;
  Tcl_Obj*              o;


  DLP_CHECKARGS (objc < 2, interp, cmd, " converter <subcommand>...");

  if (Tcl_GetIndexFromObj(interp, objv [1], subcmd, "subcommand", 0,
			  &pindex) != TCL_OK) {
    return TCL_ERROR;
  }

  if (pindex != CVT_LIST) {
    DLP_CHECKARGS (objc < 3, interp, cmd, " converter <subcommand>...");
  }

  switch (pindex) {
  case CVT_LIST:
    DLP_CHECKARGS (objc != 2, interp, cmd, " converter list");
    return PiTclPackList (interp, &info->pack);


  case CVT_CGET:
    DLP_CHECKARGS (objc != 4, interp, cmd, " converter cget <name> -option");

    if (PiTclPackFind (&info->pack, interp, objv [2], &p) != TCL_OK) {
      return TCL_ERROR;
    } else if (p == (PiTclPackInfo*) NULL) {
      Tcl_AppendResult (interp,
			"packer not found",
			(char*) NULL);
      return TCL_ERROR;
    }

    o = PiTclPackCGet (p, interp, cmd, objv [3]);

    if (o == (Tcl_Obj*) NULL) {
      return TCL_ERROR;
    } else {
      Tcl_SetObjResult (interp, o);
      return TCL_OK;
    }


  case CVT_CONF:
    if (objc == 3) {
      /* (piDlp) converter configure <name>
       * Return list of possible options.
       */

      int res;
      Tcl_Obj* list = Tcl_NewListObj (0, NULL);

      LIST_ADDSTR (list, "-appinfoPack");   LIST_ADDSTR (list, "-appinfoUnpack");
      LIST_ADDSTR (list, "-recordPack");    LIST_ADDSTR (list, "-recordUnpack");
      LIST_ADDSTR (list, "-sortinfoPack");  LIST_ADDSTR (list, "-sortinfoUnpack");

      Tcl_SetObjResult (interp, list);
      return TCL_OK;
    }

    if (PiTclPackFind (&info->pack, interp, objv [2], &p) != TCL_OK) {
      return TCL_ERROR;
    } else if (p == (PiTclPackInfo*) NULL) {
      Tcl_AppendResult (interp,
			"packer not found",
			(char*) NULL);
      return TCL_ERROR;
    }

    if (objc == 4) {
      /* (piDlp) converter configure <name> -option
       * == cget
       */

      o = PiTclPackCGet (p, interp, cmd, objv [3]);

      if (o == (Tcl_Obj*) NULL) {
	return TCL_ERROR;
      } else {
	Tcl_SetObjResult (interp, o);
	return TCL_OK;
      }
    } else {
      return PiTclPackConfigure (p, interp, cmd, objc-3, objv+3);
    }


  case CVT_CREATE:
    if (PiTclPackFind (&info->pack, interp, objv [2], &p) == TCL_OK) {
      /* Packer found, *not* desired here */

      Tcl_ResetResult  (interp);
      Tcl_AppendResult (interp,
			"packer already existent",
			(char*) NULL);
      return TCL_ERROR;
    } else if (p != (PiTclPackInfo*) NULL) {
      Tcl_ResetResult  (interp);
      Tcl_AppendResult (interp,
			"packer already existent",
			(char*) NULL);
      return TCL_ERROR;
    }

    Tcl_ResetResult  (interp);
    p = PiTclPackCreate (ARG (2));

    if (PiTclPackConfigure (p, interp, cmd, objc-3, objv+3) != TCL_OK) {
      PiTclPackRelease (p);
      return TCL_ERROR;
    }

    return PiTclPackAdd (&info->pack, interp, p);


  case CVT_DELETE:

    DLP_CHECKARGS (objc != 3, interp, cmd, " converter delete <name>");

    if (PiTclPackFind (&info->pack, interp, objv [2], &p) != TCL_OK) {
      return TCL_ERROR;
    } else if (p == (PiTclPackInfo*) NULL) {
      Tcl_AppendResult (interp,
			"packer not found",
			(char*) NULL);
      return TCL_ERROR;
    }

    return PiTclPackRemove (&info->pack, interp, p);


  default:
    return TCL_ERROR;
  }

  return TCL_ERROR;
}

/*
 *------------------------------------------------------*
 *
 *	PiTclDlpMap --
 *
 *	------------------------------------------------*
 *	This procedure realizes the 'map' subcommand
 *	of the dlp generator. See the manpages for
 *	details on what it does.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
PiTclDlpMap (interp, info, cmd, objc, objv)
Tcl_Interp* interp;
PiTclInfo*  info;
char*       cmd;
int         objc;
Tcl_Obj* CONST objv [];
{
  /* (piDlp) map add    <db> <creator> <type> <cvtName>
   * (piDlp) map remove <db> <creator> <type>
   * (piDlp) map find   <db> <creator> <type>
   *
   * (piDlp) map list
   * (piDlp) map clear
   */

  static char *subcmd[] = {
    "add", "clear", "find", "list", "remove", NULL
  };
  enum subcmd {
    MAP_ADD, MAP_CLEAR, MAP_FIND, MAP_LIST, MAP_REMOVE
  };

  PiTclMapEntry  me;
  PiTclMapEntry* mePtr;
  int            pindex;
  PiTclPackInfo* p;

  DLP_CHECKARGS (objc < 2, interp, cmd,
		 "map <subcommand>...");

  if (Tcl_GetIndexFromObj(interp, objv [1], subcmd, "subcommand", 0,
			  &pindex) != TCL_OK) {
    return TCL_ERROR;
  }

  if ((pindex != MAP_LIST) &&
      (pindex != MAP_CLEAR)) {
    DLP_CHECKARGS (objc < 5, interp, cmd,
		   "map <subcommand> <db> <creator> <type>...");
  }


  switch (pindex) {
  case MAP_LIST:
    DLP_CHECKARGS (objc != 2, interp, cmd, " map list");
    return PiTclMapList (interp, &info->map);


  case MAP_CLEAR:
    DLP_CHECKARGS (objc != 2, interp, cmd, " map clear");
    return PiTclMapClear (interp, &info->map);


  case MAP_ADD:
    DLP_CHECKARGS (objc != 6, interp, cmd,
		   "map add <db> <creator> <type> <cvt>");

    if (PiTclPackFind (&info->pack, interp, objv [5], &p) != TCL_OK) {
      return TCL_ERROR;
    } else if (p == (PiTclPackInfo*) NULL) {
      Tcl_AppendResult (interp,
			"packer not found",
			(char*) NULL);
      return TCL_ERROR;
    }

    me.database = objv [2];
    me.creator  = objv [3];
    me.type     = objv [4];
    me.pack     = p;

    return PiTclMapAdd (&info->map, &me);

  case MAP_REMOVE:
    DLP_CHECKARGS (objc > 5, interp, cmd,
		   "map remove <db> <creator> <type>");

    return PiTclMapRemove (interp, &info->map, objv [2], objv [3], objv [4]);

  case MAP_FIND:
    DLP_CHECKARGS (objc > 5, interp, cmd,
		   "map find <db> <creator> <type>");

    mePtr = PiTclMapFind (&info->map, objv [2], objv [3], objv [4]);

    if (mePtr == (PiTclMapEntry*) NULL) {
      Tcl_SetObjResult (interp, Tcl_NewStringObj ("", -1));
    } else {
      Tcl_SetObjResult (interp, Tcl_NewStringObj (mePtr->pack->name, -1));
    }
    return TCL_OK;

  default:
    return TCL_ERROR;
  }

  return TCL_ERROR;
}

/*
 *------------------------------------------------------*
 *
 *	DlpActivate --
 *
 *	------------------------------------------------*
 *	This procedure implements the 'activate' sub-
 *	command of dlp objects. See the manpages for
 *	details on what it does.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
DlpActivate (dlp, cmd, interp, argc)
PiTclDlpInstance* dlp;
char*             cmd;
Tcl_Interp*       interp;
int               argc;
{
  int sd;

  DLP_CHECKARGS (argc > 1, interp, cmd, " activate");

  if (dlp->acceptDone) {
    /* We are already in touch with the palm pilot.
     */
    return TCL_OK;
  }

  sd = pi_accept (dlp->sd, 0, 0);

  if (sd < 0) {
    Tcl_AppendResult (interp,
		      cmd, ": could not connect to pilot",
		      (char*) NULL);
    return TCL_ERROR;
  }

  dlp->sd = sd;
  dlp->acceptDone = 1;

  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	DlpDelete --
 *
 *	------------------------------------------------*
 *	This procedure realizes the 'delete' subcommand
 *	of dlp objects. See the manpages for details on
 *	what it does.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
DlpDelete (dlp, cmd, interp, argc)
PiTclDlpInstance* dlp;
char*             cmd;
Tcl_Interp*       interp;
int               argc;
{
  DLP_CHECKARGS (argc > 1, interp, cmd, " delete");

  /* The cleanup procedure will close the connection, etc.
   */

  Tcl_DeleteCommandFromToken (interp, dlp->cmd);
  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	DlpSession --
 *
 *	------------------------------------------------*
 *	This procedure realizes the 'session' subcommand
 *	of dlp objects. See the manpages for details on
 *	what it does.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
DlpSession (dlp, cmd, interp, objc, objv)
PiTclDlpInstance* dlp;
char*             cmd;
Tcl_Interp*       interp;
int               objc;
Tcl_Obj* CONST    objv [];
{
  int pindex, dlpResult;

  static char *subcmd[] = {
    "abort", "end", "log", "resetSystem", "start", NULL
  };
  enum subcmd {
    DLP_SES_ABORT, DLP_SES_END, DLP_SES_LOG, DLP_SES_RESET, DLP_SES_START
  };

  DLP_CHECKARGS (objc < 2, interp, cmd, " session <subcommand>...");

  if (Tcl_GetIndexFromObj(interp, objv [1], subcmd, "subcommand", 0,
			  &pindex) != TCL_OK) {
    return TCL_ERROR;
  }

  if (pindex == DLP_SES_LOG) {
    DLP_CHECKARGS (objc != 3, interp, cmd, " session log <text>");
  } else if (pindex == DLP_SES_END) {
    DLP_CHECKARGS ((objc < 2) || (objc > 3), interp, cmd,
		   " session end ?state?");
  } else {
    DLP_CHECKARGS_S (objc > 2, interp, cmd, " session ", ARG (1));
  }

  CHK_ACTIVE (dlp, cmd, interp);

  switch (pindex) {
  case DLP_SES_ABORT:
    dlpResult = dlp_AbortSync (dlp->sd);

    if (dlpResult >= 0) {
      /* Remember that the session is done */
      dlp->inSession = 0;
    }
    break;

  case DLP_SES_END:

    if (objc == 3) {
      static char *states [] = {
	"normal", "outofmem", "cancel", "other", NULL
      };
      
      if (Tcl_GetIndexFromObj(interp, objv [2], states, "state", 0,
				&pindex) != TCL_OK) {
	return TCL_ERROR;
      }
    } else {
      pindex = dlpEndCodeNormal;
    }

    dlpResult = dlp_EndOfSync (dlp->sd, pindex);

    if (dlpResult >= 0) {
      /* Remember that the session is done. The socket too,
       * as I was told. Recreate it, don't forget to set
       * the accept flag right.
       */

      int sd = PiTclOpenDlpChannel (dlp->dev);

      dlp->sd         = sd;
      dlp->inSession  = 0;
      dlp->acceptDone = 0;

      if (sd == -1) {
	/* Recreation failed, bail out */

	Tcl_AppendResult (interp,
			  cmd," session ",
			  ARG (1), ": ",
			  "could not recreate internal socket",
			  (char*) NULL);	
	return TCL_ERROR;
      }
    }
    break;

  case DLP_SES_LOG:
    {
      /* *HACK*
       * Experimentation showed that the last character of
       * the string is gobbled somehow. Add a dummy character
       * to get everything across the line.
       *
       * -W- Check for length limit (a message must be shorter than 64K!)
       * -W- maybe use more than one call to transfer the whole string.
       */

      int   tlen;
      char* m   = Tcl_GetStringFromObj (objv [2], &tlen);
      char* tmp = DUPN(m, 1);

      tmp [tlen]   = '.';
      tmp [tlen+1] = '\0';

      dlpResult = dlp_AddSyncLogEntry (dlp->sd, tmp);
      Tcl_Free (tmp);
    }
    break;

  case DLP_SES_RESET:
    dlpResult = dlp_ResetSystem (dlp->sd);
    break;

  case DLP_SES_START:
    dlpResult = dlp_OpenConduit (dlp->sd);

    if (dlpResult >= 0) {
      /* Remember that we have a session running now */
      dlp->inSession = 1;
    }
    break;

  default:
    /* impossible */
    return TCL_ERROR;
  }

  DLP_CHECK_S (interp, cmd, " session ", ARG(1), dlpResult);

  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	DlpSystem --
 *
 *	------------------------------------------------*
 *	This procedure realizes the 'system' subcommand
 *	of dlp objects. See the manpages for details on
 *	what it does.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
DlpSystem (dlp, cmd, interp, objc, objv)
PiTclDlpInstance* dlp;
char*             cmd;
Tcl_Interp*       interp;
int               objc;
Tcl_Obj* CONST    objv [];
{
  int pindex;

  static char *subcmd[] = {
    "card", "get", "set", NULL
  };
  enum subcmd {
    DLP_SYS_CARD, DLP_SYS_GET, DLP_SYS_SET
  };	  

  DLP_CHECKARGS (objc < 2, interp, cmd, " system <subcommand>...");

  if (Tcl_GetIndexFromObj(interp, objv [1], subcmd, "subcommand", 0,
			  &pindex) != TCL_OK) {
    return TCL_ERROR;
  }

  switch (pindex) {
  case DLP_SYS_CARD:
    return SystemCard (dlp, cmd, interp, objc-1, objv+1);

  case DLP_SYS_GET:
    return SystemGet (dlp, cmd, interp, objc-1, objv+1);

  case DLP_SYS_SET:
    return SystemSet (dlp, cmd, interp, objc-1, objv+1);

  default:
    /* impossible */
    return TCL_ERROR;
  }

  return TCL_ERROR;
}

/*
 *------------------------------------------------------*
 *
 *	DlpTickle --
 *
 *	------------------------------------------------*
 *	This procedure realizes the 'tickle' subcommand
 *	of dlp objects. See the manpages for details on
 *	what it does.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
DlpTickle (dlp, cmd, interp, objc, objv)
PiTclDlpInstance* dlp;
char*             cmd;
Tcl_Interp*       interp;
int               objc;
Tcl_Obj* CONST    objv [];
{
  int res;

  DLP_CHECKARGS (objc != 1, interp, cmd, " tickle");

  CHK_ACTIVE(dlp, cmd, interp);

  res = pi_tickle (dlp->sd);

  DLP_CHECK (interp,cmd,"tickle",res);

  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	DlpDb --
 *
 *	------------------------------------------------*
 *	This procedure realizes the 'db' subcommand of
 *	dlp objects. See the manpages for details on what
 *	it does.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
DlpDb (dlp, cmd, interp, objc, objv)
PiTclDlpInstance* dlp;
char*             cmd;
Tcl_Interp*       interp;
int               objc;
Tcl_Obj* CONST    objv [];
{
  int pindex;

  static char *subcmd[] = {
    "closeAll", "create", "delete", "info", "list", "open", NULL
  };
  enum subcmd {
    DLP_DB_CLOSEALL, DLP_DB_CREATE, DLP_DB_DELETE, DLP_DB_INFO,
    DLP_DB_LIST,     DLP_DB_OPEN
  };	  

  DLP_CHECKARGS (objc < 2, interp, cmd, " db <subcommand>...");

  if (Tcl_GetIndexFromObj(interp, objv [1], subcmd, "subcommand", 0,
			  &pindex) != TCL_OK) {
    return TCL_ERROR;
  }


  switch (pindex) {
  case DLP_DB_CLOSEALL:
    /* Close all open database objects.
     */
    DLP_CHECKARGS_S (objc > 2, interp, cmd, " db ", ARG (1));

    return PiTclDlpDbCloseAll (dlp, cmd, interp, PITCL_DB_CLOSE);

  case DLP_DB_CREATE:
    return DatabaseCreate (dlp, cmd, interp, objc-1, objv+1);

  case DLP_DB_DELETE:
    return DatabaseDelete (dlp, cmd, interp, objc-1, objv+1);

  case DLP_DB_OPEN:
    return DatabaseOpen (dlp, cmd, interp, objc-1, objv+1);

  case DLP_DB_LIST:
    DLP_CHECKARGS (objc > 2, interp, cmd, " db list");

    return DatabaseList (dlp, cmd, interp);

  case DLP_DB_INFO:
    DLP_CHECKARGS (objc > 3, interp, cmd, " db info <dbname>");

    return DatabaseInfo (dlp, cmd, interp, ARG (2));

  default:
    /* Should be impossible */
    return TCL_ERROR;
  }

  return TCL_ERROR;
}

/*
 *------------------------------------------------------*
 *
 *	SystemGet --
 *
 *	------------------------------------------------*
 *	This procedure realizes the 'system get' sub-
 *	command of dlp objects. See the manpages for
 *	details on what it does.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
SystemGet (dlp, cmd, interp, objc, objv)
PiTclDlpInstance* dlp;
char*             cmd;
Tcl_Interp*       interp;
int               objc;
Tcl_Obj* CONST    objv [];
{
  int pindex, dlpResult;

  static char *subcmd[] = {
    "info", "netsync", "time", "user", NULL
  };
  enum subcmd {
    DLP_GET_INFO, DLP_GET_NETSYNC, DLP_GET_TIME, DLP_GET_USER
  };	  

  DLP_CHECKARGS (objc != 2, interp, cmd, " system get <subcommand>");

  if (Tcl_GetIndexFromObj(interp, objv[1], subcmd, "subcommand", 0,
			  &pindex) != TCL_OK) {
    return TCL_ERROR;
  }

  CHK_ACTIVE(dlp, cmd, interp);

  switch (pindex) {
  case DLP_GET_INFO:
    {
      Tcl_Obj* res;
      SysInfo  si;

      dlpResult = dlp_ReadSysInfo (dlp->sd, &si);

      DLP_CHECK_S (interp, cmd, " system ", ARG(1), dlpResult);

      res = PiTclSysInfoToList (interp, &si);

      if (res == (Tcl_Obj*) NULL) {
	return TCL_ERROR;
      }

      Tcl_SetObjResult (interp, res);
    }
  break;

  case DLP_GET_NETSYNC:
    {
      Tcl_Obj*    res;
      NetSyncInfo nsi;

      dlpResult = dlp_ReadNetSyncInfo (dlp->sd, &nsi);

      DLP_CHECK_S (interp, cmd, " system ", ARG(1), dlpResult);

      res = PiTclNetSyncInfoToList (interp, &nsi);

      if (res == (Tcl_Obj*) NULL) {
	return TCL_ERROR;
      }

      Tcl_SetObjResult (interp, res);
    }
  break;

  case DLP_GET_TIME:
    {
      time_t pilotTime;

      dlpResult = dlp_GetSysDateTime (dlp->sd, &pilotTime);

      DLP_CHECK_S (interp, cmd, " system ", ARG(1), dlpResult);

      Tcl_SetObjResult (interp, Tcl_NewIntObj (pilotTime));
      return TCL_OK;
    }
  break;

  case DLP_GET_USER:
    {
      Tcl_Obj*  res;
      PilotUser pu;

      dlpResult = dlp_ReadUserInfo (dlp->sd, &pu);

      DLP_CHECK_S (interp, cmd, " system ", ARG(1), dlpResult);

      res = PiTclPilotUserToList (interp, &pu);

      if (res == (Tcl_Obj*) NULL) {
	return TCL_ERROR;
      }

      Tcl_SetObjResult (interp, res);
    }
  break;

  default:
    /* should be impossible to reach */
    return TCL_ERROR;
  }

  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	SystemSet --
 *
 *	------------------------------------------------*
 *	This procedure realizes the 'system card' sub-
 *	command of dlp objects. See the manpages for
 *	details on what it does.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
SystemSet (dlp, cmd, interp, objc, objv)
PiTclDlpInstance* dlp;
char*             cmd;
Tcl_Interp*       interp;
int               objc;
Tcl_Obj* CONST    objv [];
{
  int pindex, dlpResult, res;

  static char *subcmd[] = {
    "netsync", "time", "user", NULL
  };
  enum subcmd {
    DLP_SET_NETSYNC, DLP_SET_TIME, DLP_SET_USER
  };	  


  DLP_CHECKARGS (objc != 3, interp, cmd,
		 " system set <subcommand> <data>");

  if (Tcl_GetIndexFromObj(interp, objv[1], subcmd, "subcommand", 0,
			  &pindex) != TCL_OK) {
    return TCL_ERROR;
  }

  CHK_ACTIVE(dlp, cmd, interp);

  switch (pindex) {
  case DLP_SET_NETSYNC:
    {
      NetSyncInfo nsi;

      res = PiTclListToNetSyncInfo (interp, objv [2], &nsi);

      if (res != TCL_OK) {
	return res;
      }

      dlpResult = dlp_WriteNetSyncInfo (dlp->sd, &nsi);
    }
  break;

  case DLP_SET_TIME:
    {
      time_t pilotTime;

      res = Tcl_GetLongFromObj (interp, objv [2], &pilotTime);

      if (res != TCL_OK) {
	return res;
      }

      dlpResult = dlp_SetSysDateTime (dlp->sd, pilotTime);
    }
  break;

  case DLP_SET_USER:
    {
      PilotUser pu;

      res = PiTclListToPilotUser (interp, objv [2], &pu);

      if (res != TCL_OK) {
	return res;
      }

      dlpResult = dlp_WriteUserInfo (dlp->sd, &pu);
    }
  break;

  default:
    /* should be impossible to reach */
    return TCL_ERROR;
  }

  DLP_CHECK_S (interp, cmd, " system ", ARG(1), dlpResult);

  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	SystemCard --
 *
 *	------------------------------------------------*
 *	This procedure realizes the 'system card' sub-
 *	command of dlp objects. See the manpages for
 *	details on what it does.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
SystemCard (dlp, cmd, interp, objc, objv)
PiTclDlpInstance* dlp;
char*             cmd;
Tcl_Interp*       interp;
int               objc;
Tcl_Obj* CONST    objv [];
{
  int pindex;
  Tcl_Obj* res;

  static char *subcmd[] = {
    "get", "list", "num", NULL
  };
  enum subcmd {
    DLP_CARD_GET, DLP_CARD_LIST, DLP_CARD_NUM
  };	  


  DLP_CHECKARGS (objc < 2, interp, cmd,
		 " system card <subcommand>...");

  if (Tcl_GetIndexFromObj(interp, objv [1], subcmd, "subcommand", 0,
			  &pindex) != TCL_OK) {
    return TCL_ERROR;
  }

  if (pindex == DLP_CARD_GET) {
    DLP_CHECKARGS (objc != 3, interp, cmd,
		   " system card get <index>");
  } else {
    DLP_CHECKARGS_S (objc > 2, interp, cmd,
		   " system card ", ARG (1));
  }

  CHK_ACTIVE (dlp, cmd, interp);

  if (PiTclDlpReadCards (dlp, cmd, interp) != TCL_OK) {
    return TCL_ERROR;
  }

  switch (pindex) {
  case DLP_CARD_GET:

    if (Tcl_GetIntFromObj (interp, objv [2], &pindex) != TCL_OK) {
      return TCL_ERROR;
    }

    res = PiTclCardToList (interp, dlp->card [pindex]);

    if (res == (Tcl_Obj*) NULL) {
      return TCL_ERROR;
    }

    Tcl_SetObjResult (interp, res);
    break;

  case DLP_CARD_LIST:
    res = PiTclCardsToList (interp, dlp);

    if (res == (Tcl_Obj*) NULL) {
      return TCL_ERROR;
    }

    Tcl_SetObjResult (interp, res);
    break;

  case DLP_CARD_NUM:
    Tcl_SetObjResult (interp, Tcl_NewIntObj (dlp->numCards));
    break;

  default:
    /* impossible */
    return TCL_ERROR;
  }

  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	DatabaseCreate --
 *
 *	------------------------------------------------*
 *	This procedure realizes the 'db create' sub-
 *	command of dlp objects. See the manpages for
 *	details on what it does.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
DatabaseCreate (dlp, cmd, interp, objc, objv)
PiTclDlpInstance* dlp;
char*             cmd;
Tcl_Interp*       interp;
int               objc;
Tcl_Obj* CONST    objv [];
{
  /* Syntax: create <name> <creator> <type> ?option value...?
   * options:	-pack    <packerobject>
   *		-card	 <no>
   *		-flags   <flaglist>
   *		-version <version>
   */

  Tcl_HashEntry*        hPtr;
  char*                 dbName;
  char*                 dbCmdName;
  long int              dbCreator = 0;
  long int              dbType    = 0;
  PiTclDbInstance*      db;
  int                   dlpResult =  0;
  int                   dbHandle  = -1;
  int                   dbCard    =  0;
  int                   dbFlags   =  dlpOpenRead;
  int                   dbVersion =  0;
  DBInfo                dbInfo;
  PiTclPackInfo* dbPack    = (PiTclPackInfo*) NULL;


  DLP_CHECKARGS_S (objc < 4, interp, cmd, " db create ",
		   "<db> <dbname> <creator> <type> ?<option> <value>...?");

  /* Connect to pilot */

  CHK_ACTIVE (dlp, cmd, interp);

  /* Get information about *all* databases */

  if (PiTclDlpReadDbList (dlp, cmd, interp) != TCL_OK) {
    return TCL_ERROR;
  }

  dbCmdName = ARG (1);
  dbName    = ARG (2);

  hPtr = Tcl_FindHashEntry (&dlp->db, dbName);

  if (hPtr != (Tcl_HashEntry*) NULL) {
    /* Name was found, user specified already known database.
     */

    Tcl_AppendResult (interp,
		      cmd ," db create: \"",
		      dbName, "\" already existing",
		      (char*) NULL);
    return TCL_ERROR;
  }

  dbCreator = PiTclStringToLong (objv [3]);
  dbType    = PiTclStringToLong (objv [4]);

  {
    PiTclMapEntry* e = PiTclMapFind (&dlp->base.info->map,
				     objv [2], objv [3],
				     objv [4]);

    if (e != (PiTclMapEntry*) NULL) {
      dbPack = e->pack;
    }
  }

  if (ProcessCreateOptions (dlp, cmd, interp, objc-5, objv+5,
			    &dbCard, &dbFlags,
			    &dbVersion, &dbPack) != TCL_OK) {
    return TCL_ERROR;
  }

  /* Create db in the pilot, then fetch its information block.
   */

  dlpResult = dlp_CreateDB (dlp->sd, dbCreator, dbType,
			    dbCard, dbFlags, dbVersion,
			    dbName, &dbHandle);

  DLP_CHECK_S (interp, cmd, " db create ", dbName, dlpResult);

  dlpResult = dlp_FindDBInfo (dlp->sd, 0, 0, dbName, 0, 0, &dbInfo);

  DLP_CHECK_S (interp, cmd, " db create ", dbName, dlpResult);

  /* Create the new PiTclDbInstance structure and record it
   * in the local cache. Set the markers showing that
   * this instance is open, create the object command
   * for acessing it.
   */

  db = PiTclCreateDbObject (dlp, &dbInfo, dbCard);

  db->sd     = dlp->sd;
  db->handle = dbHandle;
  db->cmd    = Tcl_CreateObjCommand (interp, dbCmdName,
				     &PiTclDlpDbObjectCmd,
				     (ClientData) db,
				     &PiTclDeleteDbObjectCmd);
  db->pack   = dbPack;

  if (dbPack != (PiTclPackInfo*) NULL) {
    PiTclPackPreserve (db->pack);
  }

  Tcl_SetObjResult (interp, objv [1]);
  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	DatabaseDelete --
 *
 *	------------------------------------------------*
 *	This procedure realizes the 'db delete' sub-
 *	command of dlp objects. See the manpages for
 *	details on what it does.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
DatabaseDelete (dlp, cmd, interp, objc, objv)
PiTclDlpInstance* dlp;
char*             cmd;
Tcl_Interp*       interp;
int               objc;
Tcl_Obj* CONST    objv [];
{
  Tcl_HashEntry*   hPtr;
  char*            dbName;
  PiTclDbInstance* db;
  int              dlpResult;

  DLP_CHECKARGS (objc != 2, interp, cmd, " db delete <dbname>");

  /* Connect to pilot */

  CHK_ACTIVE (dlp, cmd, interp);

  /* Get information about *all* databases */

  if (PiTclDlpReadDbList (dlp, cmd, interp) != TCL_OK) {
    return TCL_ERROR;
  }

  /* Locate the database information and use it to destroy the object
   * command, if defined. This will close an open connection on the
   * pilot side too. Then talk to the pilot and command him to delete
   * the named DB from his memory. Only after a successful completion
   * of this step we are allowed to destroy the local db information
   * block too.
   */

  dbName = ARG (1);

  hPtr = Tcl_FindHashEntry (&dlp->db, dbName);

  if (hPtr == (Tcl_HashEntry*) NULL) {
    /* Name not found, user specified unknown database.
     */

    Tcl_AppendResult (interp,
		      cmd ," db delete: \"",
		      dbName, "\" is not known",
		      (char*) NULL);
    return TCL_ERROR;
  }

  db = (PiTclDbInstance*) Tcl_GetHashValue (hPtr);

  if ((Tcl_Command)db->cmd != (Tcl_Command) NULL) {
    Tcl_DeleteCommandFromToken (interp, db->cmd);
  }

  dlpResult = dlp_DeleteDB (dlp->sd, db->card, db->info.name);

  DLP_CHECK_S (interp, cmd, " db delete ", dbName, dlpResult);

  Tcl_DeleteHashEntry  (hPtr);
  PiTclReleaseDbObject (db);

  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	DatabaseOpen --
 *
 *	------------------------------------------------*
 *	This procedure realizes the 'db open' sub-
 *	command of dlp objects. See the manpages for
 *	details on what it does.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
DatabaseOpen (dlp, cmd, interp, objc, objv)
PiTclDlpInstance* dlp;
char*             cmd;
Tcl_Interp*       interp;
int               objc;
Tcl_Obj* CONST    objv [];
{
  /* Syntax: open <name> ?option value...?
   * options:	-pack   <packerobject>
   *		-flags	<flaglist>
   */

  Tcl_HashEntry*        hPtr;
  char*                 dbName;
  char*                 dbCmdName;
  PiTclDbInstance*      db;
  int                   dlpResult =  0;
  int                   dbHandle  = -1;
  int                   dbMode    = dlpOpenRead;
  PiTclPackInfo* dbPack    = (PiTclPackInfo*) NULL;


  DLP_CHECKARGS (objc < 3, interp, cmd,
		 " db open <db> <dbname> ?<option> <value>...?");

  /* Connect to pilot */

  CHK_ACTIVE (dlp, cmd, interp);

  /* Get information about *all* databases */

  if (PiTclDlpReadDbList (dlp, cmd, interp) != TCL_OK) {
    return TCL_ERROR;
  }

  dbCmdName = ARG (1);
  dbName    = ARG (2);

  hPtr = Tcl_FindHashEntry (&dlp->db, dbName);

  if (hPtr == (Tcl_HashEntry*) NULL) {
    /* Name not found, user specified unknown database.
     */

    Tcl_AppendResult (interp,
		      cmd ," db open: \"",
		      dbName, "\" is not known",
		      (char*) NULL);
    return TCL_ERROR;
  }

  db = (PiTclDbInstance*) Tcl_GetHashValue (hPtr);

  if (db->cmd != (Tcl_Command) NULL) {

    Tcl_AppendResult (interp,
		      cmd ," db open: \"",
		      dbName, "\" is already open",
		      (char*) NULL);

    return TCL_ERROR;
  }

  {
    PiTclMapEntry* e = PiTclMapFind (&dlp->base.info->map,
				     db->name, db->creator,
				     db->type);

    if (e != (PiTclMapEntry*) NULL) {
      dbPack = e->pack;
    }
  }

  if (ProcessOpenOptions (dlp, cmd, interp, objc-3, objv+3,
			  &dbMode, &dbPack) != TCL_OK) {
    return TCL_ERROR;
  }


  dlpResult = dlp_OpenDB (dlp->sd, db->card, dbMode,
			  dbName, &dbHandle);

  DLP_CHECK_S (interp, cmd, " db open ", dbName, dlpResult);

  /* Set the markers showing that this instance is open,
   * create the object command for acessing it.
   */

  db->cmd = Tcl_CreateObjCommand (interp, dbCmdName,
				  &PiTclDlpDbObjectCmd,
				  (ClientData) db,
				  &PiTclDeleteDbObjectCmd);

  db->sd     = dlp->sd;
  db->handle = dbHandle;
  db->pack   = dbPack;

  if (dbPack != (PiTclPackInfo*) NULL) {
    PiTclPackPreserve (db->pack);
  }

  Tcl_SetObjResult (interp, objv [1]);
  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	DatabaseList --
 *
 *	------------------------------------------------*
 *	This procedure realizes the 'db list' sub-
 *	command of dlp objects. See the manpages for
 *	details on what it does.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
DatabaseList (dlp, cmd, interp)
PiTclDlpInstance* dlp;
char*             cmd;
Tcl_Interp*       interp;
{
  /* Connect to pilot */

  CHK_ACTIVE (dlp, cmd, interp);

  /* Get information about *all* databases
   */

  if (PiTclDlpReadDbList (dlp, cmd, interp) != TCL_OK) {
    return TCL_ERROR;
  }

  /* And extract the list of names from it.
   */

  if (PiTclListDatabases (interp, dlp) !=- TCL_OK) {
    return TCL_ERROR;
  }

  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	DatabaseInfo --
 *
 *	------------------------------------------------*
 *	This procedure realizes the 'db info' sub-
 *	command of dlp objects. See the manpages for
 *	details on what it does.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
DatabaseInfo (dlp, cmd, interp, dbName)
PiTclDlpInstance* dlp;
char*             cmd;
Tcl_Interp*       interp;
char*             dbName;
{
  PiTclDbInstance* db;
  Tcl_HashEntry*   hPtr;

  /* Connect to pilot */

  CHK_ACTIVE (dlp, cmd, interp);

  /* Get information about *all* databases */

  if (PiTclDlpReadDbList (dlp, cmd, interp) != TCL_OK) {
    return TCL_ERROR;
  }

  /* Locate the database information
   */

  hPtr = Tcl_FindHashEntry (&dlp->db, dbName);

  if (hPtr == (Tcl_HashEntry*) NULL) {
    /* Name not found, user specified unknown database.
     */

    Tcl_AppendResult (interp,
		      cmd ," db info: \"",
		      dbName, "\" is not known",
		      (char*) NULL);
    return TCL_ERROR;
  }

  db = (PiTclDbInstance*) Tcl_GetHashValue (hPtr);

  if (PiTclDbObjectToList (interp, db) != TCL_OK) {
    return TCL_ERROR;
  }

  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	ProcessCreateOptions --
 *
 *	------------------------------------------------*
 *	This procedure processes the options of the
 *	subcommand 'db create'.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Sets the 'card', 'flags' and 'version'
 *		arguments to values specified by the user.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
ProcessCreateOptions (dlp, cmd, interp, objc, objv, card, flags, version, pack)
PiTclDlpInstance*      dlp;
char*                  cmd;
Tcl_Interp*            interp;
int                    objc;
Tcl_Obj* CONST         objv [];
int*                   card;
int*                   flags;
int*                   version;
PiTclPackInfo** pack;
{
  /* options:	-pack    <packerobject>
   *		-card	 <no>
   *		-flags   <flaglist>
   *		-version <version>
   */

  static char *options[] = {
    "-card", "-flags", "-pack", "-version", NULL
  };
  enum options {
    O_CARD, O_FLAGS, O_PACK, O_VERSION
  };

  int pindex;

  DLP_CHECKARGS ((objc % 2) == 1, interp, "", "an even number");

  for (; objc > 0; objc -= 2, objv += 2) {
    if (Tcl_GetIndexFromObj(interp, objv [0], options, "option", 0,
			    &pindex) != TCL_OK) {
      return TCL_ERROR;
    }

    switch (pindex) {
    case O_CARD:
      if (Tcl_GetIntFromObj (interp, objv [1], &pindex) != TCL_OK) {
	return TCL_ERROR;
      }
      *card = pindex;
      break;

    case O_VERSION:
      if (Tcl_GetIntFromObj (interp, objv [1], &pindex) != TCL_OK) {
	return TCL_ERROR;
      }
      *version = pindex;
      break;

    case O_FLAGS:
      if (PiTclEncodeDbFlags (interp, objv [1], flags) !=  TCL_OK) {
	return TCL_ERROR;
      }
      if ((*flags) & dlpDBFlagReadOnly) {
	Tcl_AppendResult (interp, "'readOnly' not allowed here, ",
			  "cannot create rom-based databases",
			  (char*) NULL);
	return TCL_ERROR;
      }
      if ((*flags) & dlpDBFlagOpen) {
	Tcl_AppendResult (interp, "'open' not allowed here, ",
			  "this flag is a read-only status indicator",
			  (char*) NULL);
	return TCL_ERROR;
      }
      break;


    case O_PACK:
      if (PiTclPackFind (&dlp->base.info->pack,
			 interp, objv [1], pack) != TCL_OK) {
	return TCL_ERROR;
      }
      break;


    default:
      /* Impossible */
      return TCL_ERROR;
    }
  }

  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	ProcessOpenOptions --
 *
 *	------------------------------------------------*
 *	This procedure processes the options of the
 *	subcommand 'db open'.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Sets the 'card', 'flags' and 'version'
 *		arguments to values specified by the user.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
ProcessOpenOptions (dlp, cmd, interp, objc, objv, mode, pack)
PiTclDlpInstance*      dlp;
char*                  cmd;
Tcl_Interp*            interp;
int                    objc;
Tcl_Obj* CONST         objv [];
int*                   mode;
PiTclPackInfo** pack;
{
  /* options:	-pack  <packerobject>
   *		-mode  <modelist>
   */

  static char *options[] = {
    "-mode", "-pack", NULL
  };
  enum options {
    O_FLAGS, O_PACK
  };

  int pindex;

  DLP_CHECKARGS ((objc % 2) == 1, interp, "", "an even number");

  for (; objc > 0; objc -= 2, objv += 2) {
    if (Tcl_GetIndexFromObj(interp, objv [0], options, "option", 0,
			    &pindex) != TCL_OK) {
      return TCL_ERROR;
    }

    switch (pindex) {
    case O_FLAGS:
      if (PiTclEncodeOpenFlags (interp, objv [1], mode) != TCL_OK) {
	return TCL_ERROR;
      }
      break;


    case O_PACK:
      if (PiTclPackFind (&dlp->base.info->pack,
			 interp, objv [1], pack) != TCL_OK) {
	return TCL_ERROR;
      }
      break;


    default:
      /* Impossible */
      return TCL_ERROR;
    }
  }

  return TCL_OK;
}
