#ifndef PITCL_MAC_H
#define PITCL_MAC_H
/* -*- c -*-
 *
 * piTclMac.h --
 *
 *	Defintion of helper and convenience macros
 *
 * Copyright (C) 1998 Andreas Kupries (a.kupries@westend.com)
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL I BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF THIS
 * SOFTWARE AND ITS DOCUMENTATION, EVEN IF I HAVE BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * I SPECIFICALLY DISCLAIM ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND
 * I HAVE NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 * ENHANCEMENTS, OR MODIFICATIONS.
 *
 * CVS: $Id: macros.h,v 1.2 1999/06/28 21:20:36 aku Exp $
 */

#ifdef __cplusplus
extern "C" {
#endif

#include <string.h>

/* Detect Tcl 8.1 and beyond
 */

#define GT81 ((TCL_MAJOR_VERSION > 8) || \
	      ((TCL_MAJOR_VERSION == 8) && \
	       (TCL_MINOR_VERSION >= 1)))


/* Convenience macros.
 * - String duplication,
 * - Bit definition,
 * - Simple access to command arguments, at least if we only
 *   need the strings stored in them.
 */

#define DUP(s)    strcpy ((char*) Tcl_Alloc (strlen((s))+1), (s))
#define DUPN(s,n) strcpy ((char*) Tcl_Alloc (strlen((s))+1+(n)), (s))
#define BIT(i)    (1<<(i))
#define ARG(i)    Tcl_GetStringFromObj (objv [i], NULL)


/* More convenience macros
 * - construction of (nested) lists
 */

#define LIST_ADDOBJ(list,o) \
  res = Tcl_ListObjAppendElement (interp, list, o); \
  if (res != TCL_OK) {                              \
    Tcl_DecrRefCount (list);                        \
    return TCL_ERROR;                               \
  }

#define LIST_ADDSTR(list,str)       LIST_ADDOBJ (list, Tcl_NewStringObj (str, -1))
#define LIST_ADDFSTR(list,str, len) LIST_ADDOBJ (list, Tcl_NewStringObj (str, len))
#define LIST_ADDINT(list,i)         LIST_ADDOBJ (list, Tcl_NewIntObj (i))


/* Construct error message from the return value of a DLP function.
 */

#define DLP_CHECK(interp,cmd,subcmd,res) \
  if ((res) < 0) { \
    Tcl_AppendResult ((interp), (cmd), " ", (subcmd), ": ", \
		      dlp_strerror (res), (char*) NULL); \
    return TCL_ERROR; \
  }

#define DLP_CHECK_S(interp,cmd,str,subcmd,res) \
  if ((res) < 0) { \
    Tcl_AppendResult ((interp), (cmd), (str), (subcmd), ": ", \
		      dlp_strerror (res), (char*) NULL); \
    return TCL_ERROR; \
  }

/* Check number of arguments, construct appropriate error message
 */

#define DLP_CHECKARGS(cond,interp,cmd,str) \
  if ((cond)) { \
    Tcl_AppendResult ((interp), "wrong # args: should be \"", \
		      (cmd) , (str), "\"", (char*) NULL); \
    return TCL_ERROR; \
  }

#define DLP_CHECKARGS_S(cond,interp,cmd,str,sub) \
  if ((cond)) { \
    Tcl_AppendResult ((interp), "wrong # args: should be \"", \
		      (cmd) , (str), (sub), "\"", (char*) NULL); \
    return TCL_ERROR; \
  }



#ifdef __cplusplus
}
#endif /* C++ */
#endif /* PITCL_MAC_H */
