#ifndef PITCL_MAP_H
#define PITCL_MAP_H
/* -*- c -*-
 *
 * map.h --
 *
 *	Global declarations of piTcl, management of packer map
 *
 * Copyright (C) 1998 Andreas Kupries (a.kupries@westend.com)
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL I BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF THIS
 * SOFTWARE AND ITS DOCUMENTATION, EVEN IF I HAVE BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * I SPECIFICALLY DISCLAIM ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND
 * I HAVE NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 * ENHANCEMENTS, OR MODIFICATIONS.
 *
 * CVS: $Id: map.h,v 1.2 1999/06/28 21:20:36 aku Exp $
 */

#ifdef __cplusplus
extern "C" {
#endif

#include <tcl.h>
#include "pack.h" /* Use packer declarations */

/* Structure of an entry in the database -> converter map
 * The patterns are glob-style (-> string match, glob).
 */

typedef struct PiTclMapEntry {
  Tcl_Obj*       database; /* database pattern to match */
  Tcl_Obj*       creator;  /* creator pattern to match */
  Tcl_Obj*       type;     /* type pattern to match */

  PiTclPackInfo* pack;     /* packer definition associated to the
			    * patterns above */
} PiTclMapEntry;


/* Structures for the management of the whole
 * database -> converter map
 */

typedef struct PiTclMapNode {
  PiTclMapEntry*       entry; /* Reference to the real entry */
  struct PiTclMapNode* next;  /* Reference to the next node in the list, NULL
			       * is used as marker for the last node in the
			       * list. */
} PiTclMapNode;


typedef struct PiTclMap {

  /* Map organization:
   * 4 lists, for patterns without, one, two or three stars in them.
   * The lists itself are sorted, with star'ed patterns behind
   * unstar'ed. Sortorder is 'database', 'creator', 'type'.
   */
  
  PiTclMapNode* list [4];  /* List of entries without, 1, 2 or 3 '*'s in
			    * their patterns */

} PiTclMap;









/* Management of map structure
 */

int
PiTclMapInit _ANSI_ARGS_ ((PiTclMap* m));

int
PiTclMapDelete _ANSI_ARGS_ ((PiTclMap* m));

int
PiTclMapAdd _ANSI_ARGS_ ((PiTclMap*      m,
			  PiTclMapEntry* e));

int
PiTclMapClear _ANSI_ARGS_ ((Tcl_Interp* interp,
			    PiTclMap*   m));

int
PiTclMapList _ANSI_ARGS_ ((Tcl_Interp* interp,
			   PiTclMap*      m));

int
PiTclMapRemove _ANSI_ARGS_ ((Tcl_Interp* interp,
			     PiTclMap*   m,
			     Tcl_Obj*    db,
			     Tcl_Obj*    creator,
			     Tcl_Obj*    type));

PiTclMapEntry*
PiTclMapFind _ANSI_ARGS_ ((PiTclMap* m,
			   Tcl_Obj*  db,
			   Tcl_Obj*  creator,
			   Tcl_Obj*  type));


#ifdef __cplusplus
}
#endif /* C++ */
#endif /* PITCL_MAP_H */
