/*
 * todo.c --
 *
 *	Definition for packer wrappers to todo databases.
 *
 * Copyright (C) 1998 Andreas Kupries (a.kupries@westend.com)
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL I BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF THIS
 * SOFTWARE AND ITS DOCUMENTATION, EVEN IF I HAVE BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * I SPECIFICALLY DISCLAIM ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND
 * I HAVE NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 * ENHANCEMENTS, OR MODIFICATIONS.
 *
 * CVS: $Id: todo.c,v 1.1 1999/03/27 13:26:34 aku Exp $
 */

#include "wrap.h"
#include "dates.h"
#include "category.h"
#include "internal.h"

#include "pi-todo.h"

typedef struct ToDo            ToDo;
typedef struct ToDoAppInfo     ToDoAppInfo;



/*
 * Definitions for conversion of todo records.
 */

static TclUtil_StructItem recordItems [] = {
  ITEM        ("priority",    priority,    ToDo, TUSI_Int),
  ITEM        ("indefinite",  indefinite,  ToDo, TUSI_Boolean),
  ITEM        ("complete",    complete,    ToDo, TUSI_Boolean),
  ITEM        ("description", description, ToDo, TUSI_String),
  ITEM        ("note",        note,        ToDo, TUSI_String),
  ITEM_CUSTOM ("due",         due,         ToDo, PiTclDateToObj, PiTclDateFromObj, NULL),
  ITEM_END
};

static TclUtil_Struct recordStruct = {
  "ToDo", 0, NULL, recordItems
};

static PiTcl_CvtData recordConverter = {
  &recordStruct,                  /* Definition of ToDo structure, for
				   * conversion to tcl list */
  sizeof (ToDo),                  /* Size of the ToDo structure */
  (PiTcl_FreeStruct*) free_ToDo,  /* Proc to free additionally allocated
				   * space */
  (PiTcl_CvtStruct*) pack_ToDo,   /* Serialize ToDo into bytestream */
  (PiTcl_CvtStruct*) unpack_ToDo, /* Unserialize bytestream into a nemo */
  4096                            /* Initial size of buffer to use for
				   * serialization */
};


/*
 * Definitions for conversion of ToDo application information.
 */

static TclUtil_EnumItem sortValues [] = {
  {"date,priority",     0},
  {"priority,date",     1},
  {"category,priority", 2},
  {"category,date",     3},
  {NULL,               -1},
};

static TclUtil_Enum sortEnum = {
  "sort mode", NULL, sortValues
};

static TclUtil_StructItem appinfoItems [] = {
  ITEM_STRUCT ("category", category,       ToDoAppInfo, &catStruct),
  ITEM_ENUM   ("sortMode", sortByPriority, ToDoAppInfo, &sortEnum),
  ITEM        ("dirty",    dirty,          ToDoAppInfo, TUSI_Boolean),
  ITEM_END
};

static TclUtil_Struct appinfoStruct = {
  "todo application information", 0, NULL, appinfoItems
};

static PiTcl_CvtData appinfoConverter = {
  &appinfoStruct,                        /* Definition of todo appinfo
					  * structure, for conversion to and
					  * from a tcl list */
  sizeof (ToDoAppInfo),                  /* Size of the todo appinfo structure
					  */
  NULL,                                  /* *No* additionally allocated space
					  */
  (PiTcl_CvtStruct*) pack_ToDoAppInfo,   /* Serialize ToDo appinfo into a
					  * bytestream */
  (PiTcl_CvtStruct*) unpack_ToDoAppInfo, /* Unserialize bytestream into a ToDo
					  * appinfo block */
  sizeof (ToDoAppInfo)+1                 /* Initial size of buffer to use for
					  * serialization */
};

/*
 * ToDo databases do not have a sort information block.
 */

/*
 * Definition of the whole ToDo converter.
 */

static PiTcl_PilotCvt converter = {
  &recordConverter,
  NULL,
  &appinfoConverter
};

/*
 *------------------------------------------------------*
 *
 *	PiTclInitToDo --
 *
 *	------------------------------------------------*
 *	This procedure registers the commands wrapped
 *	around the (un)pack functionality for ToDos.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

int
PiTclInitTodo (interp)
Tcl_Interp* interp;
{
  Tcl_CreateObjCommand (interp, "piTodo",
			&PiTclPilotCvtCmd,
			(ClientData) &converter,
			NULL);

  return TCL_OK;
}

